from . import text

def ErrorBox(error: Exception) -> None:
    '''Shows the type and message of an error on the console.
    
    error paramter must be an instance of Exception class or one of its sub-classes.'''
    
    # Checking whether error is an exception...
    try:
        if not issubclass(error.__class__ ,Exception):
            raise TypeError()
    except:
        raise TypeError()
    
    # Getting error type...
    error_type = ''
    if hasattr(error, '__module__'):
        error_type = error.__module__ + '.'
    error_type += error.__class__.__name__
    
    # Getting error message...
    error_msg  = str(error)

    sqz_msg_iter = text.squeeze_text_iter(error_msg, 56)

    print()
    # Drawing a box using unicode box-drawing characters & printing the type of the error inside it...
    print(' ' * 10, '┌', '─' * 58, '┐', sep='')
    print(' ' * 10, '│ ', sep='', end='')
    print(f'{error_type:<56}', end='')
    print(' │')
    print(' ' * 10, '├', '─' * 58, '┤', sep='')
    # Printing the message of the error...
    print(' ' * 10, '│', ' ' * 58, '│', sep='')
    for line in sqz_msg_iter:
        print(' ' * 10, '│ ', f'{line:<56}', ' │', sep='')
    print(' ' * 10, '│', ' ' * 58, '│', sep='')
    print(' ' * 10, '└', '─' * 58, '┘', sep='')


from enum import Enum
import string
from typing import Iterator, Sequence, Iterable

from .exceptions import LoopBreakException


# Defining constants...
class Delimiters(Enum):
    WHITE_SPACES = string.whitespace
    NEW_LINE_CHARS = [
        '\n',	    # Line Feed
        '\r',	    # Carriage Return
        '\r\n',	    # Carriage Return + Line Feed
        '\v',	    # Line Tabulation
        '\f',	    # Form Feed
        '\x1c',	    # File Separator
        '\x1d',	    # Group Separator
        '\x1e', 	# Record Separator
        '\x85',	    # Next Line (C1 Control Code)
        '\u2028',	# Line Separator
        '\u2029',]	# Paragraph Separator


def IndexAnyLTR(
        text: str,
        search: str,
        start: int | None = None,
        end: int | None = None,
        /
        ) -> int:
    """Returns the first index of occurrence any of characters of 'search'
    in 'text' from left ('start' defaults to None or 0 index) to right ('end'
    defaults to None or len(text) - 1).

    Exceptions:
    TypeError: there are some invalid parameter types or 'end' is less than
    'start'.
    ValueError: there is no occurrence of 'search' characters from 'start'
    index in the 'text'.
    """
    from functools import partial

    # Checking parameters types...
    if not isinstance(text, str):
        raise TypeError("'text' must be string.")
    if not isinstance(search, str):
        raise TypeError("'search' must be string.")
    if (start is not None) and (not isinstance(start, int)):
        raise TypeError("'start' must be None or integer.")
    if (end is not None) and (not isinstance(end, int)):
        raise TypeError("'end' must be None or integer.")

    def _GetIndexLTR(
            ch: str,
            text: str,
            start: int | None = None,
            end: int | None = None
            ) -> int | None:
        try:
            return text.index(ch, start, end)
        except ValueError:
            return None

    FindAnyLTR = partial(
        _GetIndexLTR,
        text=text,
        start=start,
        end=end)
    try:
        return min(filter(
            lambda a: a is not None,
            map(
                FindAnyLTR,
                search)))
    except ValueError as err:
        err.args = (
            "No occurrence of any of 'search' was found in 'text'",)
        raise err


def IndexAnyRTL(
        text: str,
        search: str,
        start: int | None = None,
        end: int | None = None,
        /
        ) -> int:
    """Returns the first index of occurrence any of characters of 'search'
    in 'text' from right ('end' defaults to None or len(text) - 1) to  left
    ('start' defaults to None or 0 index) .

    Exceptions:
    TypeError: there are some invalid parameter types or 'end' is less than
    'start'.
    ValueError: there is no occurrence of 'search' characters from 'start'
    index in the 'text'.
    """
    from functools import partial

    # Checking parameters types...
    if not isinstance(text, str):
        raise TypeError("'text' must be string.")
    if not isinstance(search, str):
        raise TypeError("'search' must be string.")
    if (start is not None) and (not isinstance(start, int)):
        raise TypeError("'start' must be None or integer.")
    if (end is not None) and (not isinstance(end, int)):
        raise TypeError("'end' must be None or integer.")

    def _GetIndexRTL(
            ch: str,
            text: str,
            start: int | None = None,
            end: int | None = None
            ) -> int | None:
        try:
            return text.rindex(ch, start, end)
        except ValueError:
            return None

    FindAnyRTL = partial(
        _GetIndexRTL,
        text=text,
        start=start,
        end=end)
    try:
        return max(filter(
            lambda a: a is not None,
            map(
                FindAnyRTL,
                search)))
    except ValueError as err:
        err.args = (
            "No occurrence of any of 'search' was found in 'text'",)
        raise err


def SplitAnyIter(
        text: str,
        delimiters: str | Delimiters = Delimiters.WHITE_SPACES,
        minSize: int = 0
        ) -> Iterator[str]:
    """Splits 'text' at any element of 'delimiters' and returns them as
    iterator (not a list). With the specified 'minSize' this slicing
    happens after that number of characters. For delimiters any member
    of Delimiters enumeration in this module can be used, or you can
    assign any string, default is Delimiters.WHITE_SPACES. This function
    'eats' any delimiter at the start of chuncks.
    """
    if isinstance(delimiters, Delimiters):
        delimiters = delimiters.value
    start = 0
    while True:
        # Stripping (eating) leading delimiters at the start of this chunck...
        try:
            while text[start] in delimiters:
                start += 1
        except IndexError:
            # There were only delimiters on the remainder of 'text'...
            # Returning an empty string...
            yield ''
            return

        end = start + minSize
        if minSize:
            if end >= len(text):
                yield text[start:]
                return
            elif text[end] in delimiters:
                yield text[start:end]
                start = end
                continue

        # Finding the nearest delimiter on the right...
        try:
            delimIndex = IndexAnyLTR(
                text,
                delimiters,
                end)
        except ValueError:
            pass
        if delimIndex:
            yield text[start:delimIndex]
            start = delimIndex
        else:
            # No white delimiter was found on the remainder of 'text'
            # Returning the rest of text
            yield text[start:]
            return


def GetCommonAffix(
        texts: Iterable[Sequence],
        is_suffix: bool = False
        ) -> slice:
    '''Returns the common affix, either prefix or suffix, of two or more
    sequences and returns a slice object specifying the intersection (at
    the start or end). It accepts two or more sequences, aggregated in
    'texts' parameter, if less is provided TypeError exception will be
    raised. The optional 'is_suffix' parameter specifies the affix, either
    prefix or suffix. So to find the common suffix set this parameter
    to true.
    '''

    # Checking parameters...
    try:
        iter(texts)
    except TypeError:
        raise("'texts' parameter must be an ietrable.")

    if len(texts) < 2:
        raise TypeError('At least two sequences must be provided.')

    if is_suffix:
        startIndex = -1
        increment = -1
    else:
        startIndex = 0
        increment = 1

    index = startIndex
    while True:
        try:
            for seqIndex in range(len(texts) - 1):
                if texts[seqIndex][index] != texts[seqIndex + 1][index]:
                    # Stopping comparisons
                    # when the first inequality if found...
                    raise LoopBreakException
        except (IndexError, LoopBreakException):
            # Stopping comparisons when the first sequence is exhausted...
            break

        index += increment

    if is_suffix:
        return slice(index - startIndex, None)
    else:
        return slice(startIndex, index)


def squeeze_text(text, sqz_width, /, break_width=10) -> list:
    '''Accepts a text and returns a list of strings of specified squeezed width.
    
    break_width specifies how many characters must search for word wrap otherwise uses hyphen to break up the last word in each line (each string element in the list return value).'''
    # Checking the correctness of parameters...
    if not isinstance(text, str):
        raise TypeError("'text' argument must be a string")
    if not isinstance(sqz_width, int):
        raise TypeError("'sqz_width' argument must be an integer")
    if sqz_width < 3:
        raise ValueError("'sqz_width' must be greater than 2")
    if not isinstance(break_width, int):
        raise TypeError("'break_width' argument must be an integer")
    if break_width < 1:
        raise ValueError("'sqz_width' must be at least 1")
    
    NEW_LINE = '\n'
    WHITE_SPACE = ' '
    lines = []
    start_index = 0
    next_new_line = text.find(NEW_LINE)
    while True:
        # Ensuring that
        # Firstly NEW_LINE exists (-1 < next_new_line)
        # Secondly lies in the scope of current line, less than sqz_width away (next_new_line < start_index + sqz_width)
        if (-1 < next_new_line < start_index + sqz_width):
            # Then append a slice of text from start_index to next_new_line to the result (including the former & excluding the latter)
            lines.append(text[start_index:next_new_line])
            start_index = next_new_line + 1
            next_new_line = text.find(NEW_LINE, start_index)
        # Else either of the above conditions is not satisfied
        # First of all, checking we are at the end of passed-in string (text)...
        elif (len(text) < start_index + sqz_width):
            # If so, adding the rest of the passed-in string to the result...
            lines.append(text[start_index:])
            break
        # Else
        # Second of all, checking there is a WHITE_SPACE just after this slice...
        elif (text[start_index + sqz_width] == WHITE_SPACE):
            # If it is the case, adding a slice from text with the length of sqz_width to the result & starting next slice just after it...
            lines.append(text[start_index:start_index + sqz_width])
            start_index = start_index + sqz_width + 1
        else:
            # And third of all, checking there is a WHITE_SPACE in the breaking zone of the current slice (in the break_width characters at the end of this slice)...
            ws_index = text.rfind(
                WHITE_SPACE,
                start_index + sqz_width - break_width,
                start_index + sqz_width
            )
            # If there is, breaking up the current line at that position...
            if (ws_index > -1):
                lines.append(text[start_index:ws_index])
                start_index = ws_index + 1
            else:
                # If there is not, move the last character of this slice tothenext slice & replace it with dash (-) character...
                lines.append(text[start_index:start_index + sqz_width - 1] + '-')
                start_index += (sqz_width - 1)
    
    return lines

def squeeze_text_iter(text, sqz_width, /, break_width=10):
    '''Accepts a text and yields a list of strings of specified squeezed width.
    
    break_width specifies how many characters must search for word wrap otherwise uses hyphen to break up the last word in each line (each string element in the list return value).'''
    NEW_LINE = '\n'
    WHITE_SPACE = ' '
    start_index = 0
    next_new_line = text.find(NEW_LINE)
    while True:
        # Ensuring that
        # Firstly NEW_LINE exists (-1 < next_new_line)
        # Secondly lies in the scope of current line, less than sqz_width away (next_new_line < start_index + sqz_width)
        if (-1 < next_new_line < start_index + sqz_width):
            # Then yielding a slice of text from start_index to next_new_line to the result (including the former & excluding the latter)
            yield text[start_index:next_new_line]
            start_index = next_new_line + 1
            next_new_line = text.find(NEW_LINE, start_index)
        # Else either of the above conditions is not satisfied
        # First of all, checking we are at the end of passed-in string (text)...
        elif (len(text) < start_index + sqz_width):
            # If so, yielding the rest of the passed-in string to the result...
            yield text[start_index:]
            break
        # Else
        # Second of all, checking there is a WHITE_SPACE just after this slice...
        elif (text[start_index + sqz_width] == WHITE_SPACE):
            # If it is the case, yielding a slice from text with the length of sqz_width to the result & starting next slice just after it...
            yield text[start_index:start_index + sqz_width]
            start_index = start_index + sqz_width + 1
        else:
            # And third of all, checking there is a WHITE_SPACE in the breaking zone of the current slice (in the break_width characters at the end of this slice)...
            ws_index = text.rfind(
                WHITE_SPACE,
                start_index + sqz_width - break_width,
                start_index + sqz_width
            )
            # If there is, breaking up the current line at that position...
            if (ws_index > -1):
                yield text[start_index:ws_index]
                start_index = ws_index + 1
            else:
                # If there is not, move the last character of this slice to the next slice & replace it with dash (-) character...
                yield text[start_index:start_index + sqz_width - 1] + '-'
                start_index += (sqz_width - 1)


import winreg

def installed_fonts_str() -> tuple:
    """Retruns a list of string which every element is a string consists of family name and also style if it has."""
    
    # Connecting to the registry of the local computer...
    hklm = winreg.ConnectRegistry(
        None,
        winreg.HKEY_LOCAL_MACHINE
    )
    # Getting installed fonts registry key which is a subkey of HKEY_LOCAL_MACHINE...
    fonts_regkey = winreg.OpenKey(
        hklm,
        "software\\microsoft\\windows nt\\currentversion\\fonts"
    )
    fonts=[]
    i=0
    while True:
        try:
            vname, value, vtype = winreg.EnumValue(fonts_regkey, i)
            if (vname[-1] == ")"):
                vname = vname[0:(vname.rindex("(") - 1)]
            fonts.append(vname)
            i += 1
        except WindowsError:
            break
    fonts_regkey.Close()
    
    return fonts


"""This module exposes some useful APIs to work with and extend traditional
console programming.
"""

from threading import Thread, Event
from time import sleep


class WaitPromptingThrd(Thread):
    """Instances of this class wrtie a waiting message while an operation
    is performing in another thread. As long as an instance of this class
    is alive, avoid wrtiting to the console. Typically you have to create
    an instance passing the 'event' argument at least, call 'start' method,
    and set the event when waiting is no longer needed then call 'join'
    method, meanwhile this object writes a waiting message to the console.
    """
    def __init__(
            self,
            event: Event,
            *,
            n: int = 3,
            waitMsg: str = 'Waiting',
            endMsg: str = '',
            daemon: bool | None = None) -> None:
        """Initializes an instance of this class.

        'event', of type threading.Event, specifies when to quit this thread.
        'n' specifies number of dots in waiting message.
        'waitMsg' specifies the message while waiting, defaults to
        'Waiting'. 'endMsg' specifies the message when event is set.
        """
        super().__init__(name='WaitMsgThrd', daemon=daemon)
        self._n = n
        """Specifies number of dots in waiting message."""
        self._waitMsg = waitMsg
        """Specifies the message while waiting."""
        self._endMsg = endMsg
        """Specifies the message when event is set."""
        self._event = event
        """Specifies when to quit this thread."""
    
    def run(self) -> None:
        maxMsgLen = len(self._waitMsg) + self._n
        while True:
            print(end='\r', flush=True)
            print(' ' * maxMsgLen, end='\r', flush=True)
            print(self._waitMsg, end='', flush=True)
            sleep(0.5)
            if self._event.is_set():
                break
            for idx in range(self._n):
                print('.', end='', flush=True)
                sleep(0.5)
                if self._event.is_set():
                    break
            else:
                continue
            break
        print(end='\r', flush=True)
        print(' ' * maxMsgLen, end='\r', flush=True)
        print(self._endMsg, end='', flush=True)



from enum import IntEnum, IntFlag


class Modifiers(IntFlag):
    """Specifies modifiers for Key event. The 'state' attribute of Key event
    is an integer (actually a bit mask) which every bit specifies a modifier
    such as Alt, Num Lock, Button-1 (left mouse button) ans so on.
    """
    SHIFT = 0x0001
    CAPS_LOCK = 0x0002
    CONTROL = 0x0004
    NUM_LOCK = 0x0008
    SCROLL_LOCK = 0x0020
    BUTTON_1 = 0x0100
    BUTTON_2 = 0x0200
    BUTTON_3 = 0x0400
    ALT = 0x20000


class KeyCodes(IntEnum):
    DELETE = 46
    A = 65
    B = 66
    I = 73
    K = 75
    L = 76
    N = 78
    O = 79
    P = 80
    S = 83
    U = 85
    V = 86
    F5 = 116


"""This module offers the following complementary exceptions:

⬤ LoopBreakException
⬤ DuplicateError
⬤ InsertionError
⬤ IntervalError
"""

class LoopBreakException(Exception):
    '''This exception is used to break nested loops.'''
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class DuplicateError(Exception):
    '''This exception is used to notify duplicate and redundancy situations.'''
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class InsertionError(Exception):
    '''This exception is used to specify a situation where an insertion operation is impossible either logically or technically.'''
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class IntervalError(Exception):
    '''This exception is used to specify an improper interval or an operation which cannot be done in the specified interval.'''
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



# Copyright (c) 2022, Megacodist
# All rights reserved.
#
# This source code is licensed under the MIT license found in the LICENSE
# file in the root directory of this source tree.

# Requires Python 3.10+
# For postponed evaluation of annotations, required CPython 3.8+
from __future__ import annotations
from bisect import bisect_right
from collections.abc import Sequence, Iterator, Iterable
from copy import deepcopy
from enum import IntEnum
from typing import Any, Callable

from ..exceptions import IntervalError


class CollisionPolicy(IntEnum):
    '''This enumeration is used when you want to add a value into a sorted
    list which is already in the list (collision). Possible policies are as
    follow:
    
    ⬤ IGNORE: specifies not to put that value into the sorted list.
    ⬤ END: specifies to put that value just after all the same values
    ⬤ START: specifies to put that value just before all the same values
    '''
    IGNORE = 0
    END = 1
    START = 2


class _DataComparerPair(object):
    def __init__(
            self,
            data: Any,
            comparer: Any
            ) -> None:
        self.data = data
        self.comparer = comparer

    def __gt__(self, value: Any) -> bool:
        return self.comparer > value
    
    def __lt__(self, value: Any) -> bool:
        return self.comparer < value
    
    def __ge__(self, value: Any) -> bool:
        return self.comparer >= value
    
    def __le__(self, value: Any) -> bool:
        return self.comparer <= value
    
    def __eq__(self, value: Any) -> bool:
        return self.comparer == value
    
    def __ne__(self, value: Any) -> bool:
        return self.comparer != value


class SortedList(Sequence):
    '''SortedList is a sequence-like, iterable, and iterator class. This
    class is not thread safe.
    '''

    def __init__(
            self,
            *,
            cp: CollisionPolicy = CollisionPolicy.IGNORE,
            key: Callable[[Any], Any] = None
            ) -> None:
        '''This is the initializer of the class with two keyword-only
        arguments. The 'cp' parameter accepts a CollisionPolicy which
        defaults to IGNORE. It can be set at any time with 'cp' property.
        The 'key' callable acts like key parameters in Python ecosystem. The
        callable must accepts a value and returns a result as well and if
        it does not provide values will be compared directly without any
        intermediate comparers. This attribute can not be changed. To change
        this attribute, you must instantiate a new sorted list.
        '''
        # Setting the comparer callable...
        self._key: Callable[[Any], Any] = key
        
        # Checking the collision policy...
        if not isinstance(cp, CollisionPolicy):
            raise TypeError("'collision' must be a " + CollisionPolicy.__name__)
        # No error, setting it...
        self._cp: CollisionPolicy = cp

        # Creating the iterator index...
        self._iterIndex: (None | int) = None

        # Initializing the internal list to empty...
        self._items: list[Any] = []
    
    @property
    def items(self) -> list[Any]:
        """Gets or sets the items of this object. It returns a copy of the
        objects in the list as a regular list or sets the underlying list
        with the the sorted version of argument provided.
        """
        return deepcopy(self._items)
    
    @items.setter
    def items(self, __list: list[Any], /) -> None:
        del self._items
        self._items = []
        self.merge(__list)
    
    @property
    def cp(self) -> CollisionPolicy:
        """Gets or sets the collision policy for this object."""
        return self._cp
    
    @cp.setter
    def cp(self, __cp: CollisionPolicy) -> None:
        if not isinstance(__cp, CollisionPolicy):
            raise TypeError(
                "The argument must be an instance of CollisionPolicy")

    def __len__(self) -> int:
        return len(self._items)
    
    def __getitem__(self, __idx: int, /) -> Any:
        if self._key:
            return self._items[__idx].data
        else:
            return self._items[__idx]
    
    def __delitem__(self, __idx: int, /) -> None:
        """Deletes the specified list position (index)."""
        del self._items[__idx]
    
    def __eq__(self, __list: list, /) -> bool:
        return self._items == __list
    
    def __ne__(self, __list: list, /) -> bool:
        return self._items != __list
    
    def __gt__(self, __list: list, /) -> bool:
        return self._items > __list
    
    def __ge__(self, __list: list, /) -> bool:
        return self._items >= __list
    
    def __lt__(self, __list: list, /) -> bool:
        return self._items < __list
    
    def __le__(self, __list: list, /) -> bool:
        return self._items <= __list

    def __iter__(self) -> Iterator[SortedList]:
        # Settingand returning the iterator variable...
        self._iterIndex = 0
        return self
    
    def __next__(self):
        try:
            currIndex = self._iterIndex
            self._iterIndex += 1
            if self._key:
                return self._items[currIndex].data
            else:
                return self._items[currIndex]
        except IndexError:
            raise StopIteration
    
    def __str__(self) -> str:
        if self._key:
            return str([item.data for item in self._items])
        else:
            return str(self._items)
    
    def __contains__(self, value: object) -> bool:
        """Determines that specified value exists in the list or not."""
        if self._key:
            value_ = _DataComparerPair(
                value,
                self._key(value))
        else:
            value_ = value
        existed, _ = self.index(value_)
        return existed
    
    def count(self, __value: Any, /) -> int:
        '''Returns number of occurrences of '__value' in the SortedList.'''
        existed, idx = self.index(__value)
        if existed:
            if isinstance(idx, slice):
                lower, upper, _ = idx.indices(self.__len__())
                return upper - lower
            else:
                return 1
        else:
            return 0
    
    def index(
            self,
            value: Any,
            start: int = 0,
            end: int | None = None
            ) -> tuple[bool, int | slice]:
        '''This method is the backbone of this class. It returns the index of
        'value' in the SortedList. You can specifies an interval in the form of
        [start, end)  to search for index, 'start' is included and 'end' in
        excluded. 'start' and 'end' are defaulted to 0 and None respectively
        which means 0 <= index < len(internal list). If you specify each
        boundary beyond these boundaries, they automatically will be clipped
        to these deafults. If there are the same value at the boundary of the
        specified interval, this method returns an index or a slice object
        that might lie outside of the interval.

        It returns a pair, the first element specifies whether value existed
        in the list (True if existed). If existed (True), the second value
        will be an integer or a slice object. If there is one such value, the
        second element will be its index in the list, or if there are multiple
        such values the second element will be a slice object. If not existed
        (False) the second element will be the proper insertion position of
        the value in the list.
        
        Exceptions:
        ⬤ TypeError: at least one of the 'start' or 'end' parameters got
        improper value
        ⬤ ValueError: 'start' was evaluated to be greater than 'end' or
        'value' is not compatible with 'key' comparer
        ⬤ IntervalError: it is impossible to put 'value' in this interval
        [start, end)'''

        # Checking boundaries & initializing left & right indexes...
        # Checking 'start' parameter...
        if not isinstance(start, int):
            raise TypeError("'start' must be an integer")
        # Clipping out of boundary indices for 'start'...
        if start < 0:
            start = 0
        
        # Checking 'end' parameter...
        if end is None:
            end = len(self._items)
        elif not isinstance(end, int):
            raise TypeError("'end' must be an integer or None")
        else:
            # Clipping out of boundary indices for 'end'...
            if end > len(self._items):
                end = len(self._items)
        
        # Checking the interval...
        if start > end:
            raise ValueError("'start' was evaluated to be greater than 'end'")
        
        # Getting comparer...
        try:
            if self._key and not isinstance(value, _DataComparerPair):
                value_ = _DataComparerPair(
                    value,
                    self._key(value))
            else:
                value_ = value
        except Exception as err:
            raise ValueError("'value' is not compatible with 'key' comparer.")
        
        # Finding index...
        idx = bisect_right(self._items, value_, start, end)
        if idx == 0:
            return False, 0
        else:
            if self._items[idx - 1] > value_:
                raise IntervalError(
                    "Impossible to add 'value' at the specified interval")
            elif self._items[idx - 1] == value_:
                # Looking for the start of the possible slice...
                start = idx - 1
                while start >= 0:
                    if self._items[start] != value_:
                        break
                    start -= 1
                start += 1
                if (idx - start) > 1:
                    return True, slice(start, idx)
                else:
                    return True, start
            else:
                return False, idx
    
    def add(
            self,
            value: Any,
            start: int = 0,
            end: int | None = None,
            cp: CollisionPolicy | None = None
            ) -> None | int:
        '''Adds 'value' into its correct position in the SortedList. It
        accepts a 'cp' parameter which can be any CollisionPolicy value
        and defaults to None. None means use object default CollisionPolicy
        or you can specifies the policy for this add operation. This method
        returns the insertion position as an integer or None if IGNORE
        CollisionPolicy prevented the insertion. You can specify an interval
        like 'index' method.
        '''
        # Determining the collision policy for this operation...
        if not ((cp is None) or isinstance(cp, CollisionPolicy)):
            raise TypeError("'cp' must be a CollisionPolicy")
        if cp is None:
            cp = self._cp
        
        # Getting comparer...
        try:
            if self._key:
                value_ = _DataComparerPair(
                    value,
                    self._key(value))
            else:
                value_ = value
        except Exception as err:
            raise ValueError("'value' is not compatible with 'key' comparer.")

        existed, idx = self.index(value_, start, end)
        if existed:
            if cp == CollisionPolicy.IGNORE:
                return None
            if isinstance(idx, slice):
                lower, upper, _ = idx.indices(self.__len__())
            else:
                # idx is an integer
                lower, upper = idx, idx + 1
            if cp == CollisionPolicy.END:
                self._items.insert(upper, value_)
                return upper
            else:
                # cp == CollisionPolicy.START
                self._items.insert(lower, value_)
                return lower
        else:
            self._items.insert(idx, value_)
            return idx
    
    def merge(
            self,
            items: Iterable[Any],
            cp: CollisionPolicy | None = None
            ) -> None:
        """Merges an interable of items into their suitable positions in
        the list. You can specify a custom collision policy for this
        operation.
        """
        # Determining the collision policy for this operation...
        if not ((cp is None) or isinstance(cp, CollisionPolicy)):
            raise TypeError("'cp' must be a CollisionPolicy")
        if cp is None:
            cp = self._cp
        
        # Preparing the iterable...
        if self._key:
            items_ = [
                _DataComparerPair(
                    data=item,
                    comparer=self._key(item))
                for item in items]
        else:
            items_ = list(items)
        items_.sort()
        
        # Inserting the iterable into the list...
        start = 0
        for item in items_:
            existed, idx = self.index(item, start)
            if existed:
                if cp != CollisionPolicy.IGNORE:
                    if isinstance(idx, slice):
                        lower, upper, _ = idx.indices(self.__len__())
                    else:
                        lower, upper = idx, idx + 1
                    if cp == CollisionPolicy.END:
                        self._items.insert(upper, item)
                        start = upper
                    else:
                        self._items.insert(lower, item)
                        start = lower
            else:
                self._items.insert(idx, item)
                start = idx




class SingletonMeta(type):
    """Apply this meta clss to any class that you want to be singleton
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]



from collections import deque
from html.parser import HTMLParser


_SINGLETON_ELEMS = [
    'area',
    'base',
    'br',
    'col',
    'command',
    'embed',
    'hr',
    'img',
    'input',
    'keygen',
    'link',
    'meta',
    'param',
    'source',
    'track',
    'wbr']


class TagMismatchError(Exception):
    """This exception is raised if the elements of an HTML or XML
    code snippet does not match.
    """
    pass


class HtmlTagsChecker(HTMLParser):
    '''An instance of this class can be used to check thecorrect layout
    of tags in an HTML code. The 'feed' method is the only method
    that should be used.
    '''
    def __init__(self, *, convert_charrefs: bool = True) -> None:
        super().__init__(convert_charrefs=convert_charrefs)

        # Definning instance attributes...
        self._tagsStack: list[str] = []

    def feed(self, data: str) -> bool:
        '''Determines that the provided data (HTML code) has
        well-formed tag structure or not.
        '''
        try:
            super().feed(data)
            if len(self._tagsStack) == 0:
                return True
            else:
                return False
        except TagMismatchError:
            return False

    def handle_decl(self, decl: str) -> None:
        super().handle_decl(decl)

    def handle_starttag(
            self,
            tag: str,
            attrs: list[tuple[str, str | None]]
            ) -> None:

        self._tagsStack.append(tag)

    def handle_endtag(self, tag: str) -> None:
        try:
            while True:
                if (tag == self._tagsStack[-1]):
                    # Its corresponding starting tag is found...
                    self._tagsStack.pop()
                    return
                elif (self._tagsStack[-1] in _SINGLETON_ELEMS):
                    self._tagsStack.pop()
                # So far we have not found a corresponding starting tag for it
                elif tag in _SINGLETON_ELEMS:
                    return
                else:
                    raise TagMismatchError()
        except IndexError:
            # Tags stack is exhausted, so it's an orphan ending tag...
            raise TagMismatchError()


class HtmlTagsOutliner_backup(HTMLParser):
    '''An instance of this class can be used to get a beautiful
    outline of elements (only elements) of tags in an HTML code.
    The 'feed' method is the only method that should be used.
    '''
    def __init__(
            self,
            indent: int = 4,
            *,
            convert_charrefs: bool = True
            ) -> None:

        super().__init__(convert_charrefs=convert_charrefs)

        # Definning instance attributes...
        self._level = 0
        self._tagsStack: deque[str] = deque()
        self._indent = ' ' * indent
        self._output = ''
        self._lastTag = ''
        self._isLastOpen = False

    def feed(self, data: str) -> str:
        '''Returns a beautiful outline of elements (only elements) in the
        provided data (HTML code).
        '''
        super().feed(data)
        if not len(self._tagsStack):
            return self._output
        else:
            raise TagMismatchError()

    def _append_tag(
            self,
            tag: str
            ) -> None:
        pre = ''
        try:
            lastOpenShevron = self._output.rindex('<')
            lastTag = self._output[lastOpenShevron:-1]
            if lastTag != tag:
                pre = '\n' + (self._indent * self._level)
        except ValueError:
            pass
        self._output += (pre + tag)

    def handle_decl(self, decl: str) -> None:
        self._append_tag(f'<!{decl}>')

    def unknown_decl(self, data: str) -> None:
        self._append_tag(f'<?{data}>')

    def handle_starttag(
            self,
            tag: str,
            attrs: list[tuple[str, str | None]]
            ) -> None:

        self._append_tag(f'<{tag}>')
        self._tagsStack.append(tag)
        self._level += 1
        self._lastTag = tag
        self._isLastOpen = True

    def handle_endtag(self, tag: str) -> None:
        try:
            while True:
                if (tag == self._tagsStack[-1]):
                    # Its corresponding starting tag is found...
                    self._append_tag(f'</{tag}>', next_line=False)
                    self._tagsStack.pop()
                    self._level -= 1
                    return
                elif (self._tagsStack[-1] in _SINGLETON_ELEMS):
                    self._append_tag(f'</{tag}>')
                    self._tagsStack.pop()
                    self._level -= 1
                # So far we have not found a corresponding starting tag for it
                elif tag in _SINGLETON_ELEMS:
                    return
                else:
                    raise TagMismatchError()
        except IndexError:
            # Tags stack is exhausted, so it's an orphan ending tag...
            raise TagMismatchError()


class HtmlTagsOutliner(HTMLParser):
    '''An instance of this class can be used to get a beautiful
    outline of elements (only elements) of tags in an HTML code.
    The 'feed' method is the only method that should be used.
    '''
    def __init__(
            self,
            indent: int = 4,
            *,
            convert_charrefs: bool = True
            ) -> None:

        super().__init__(convert_charrefs=convert_charrefs)

        # Definning instance attributes...
        self._level = 0
        self._tagsStack: deque[str] = deque()
        self._indent = ' ' * indent
        self._output = ''
        self._lastTag = ''
        self._isLastOpen = False

    def feed(self, data: str) -> str:
        '''Returns a beautiful outline of elements (only elements) in the
        provided data (HTML code).
        '''
        super().feed(data)
        if not len(self._tagsStack):
            return self._output
        else:
            raise TagMismatchError()

    def _append_tag(
            self,
            tag: str,
            next_line: bool = True
            ) -> None:

        if next_line:
            pre = '\n' + (self._indent * self._level)
        else:
            pre = ''
        self._output += (pre + tag)

    def handle_decl(self, decl: str) -> None:
        self._append_tag(f'<!{decl}>')

    def unknown_decl(self, data: str) -> None:
        self._append_tag(f'<?{data}>')

    def handle_starttag(
            self,
            tag: str,
            attrs: list[tuple[str, str | None]]
            ) -> None:

        self._append_tag(f'<{tag}>', next_line=True)
        self._tagsStack.append(tag)
        self._level += 1
        self._lastTag = tag
        self._isLastOpen = True

    def handle_endtag(self, tag: str) -> None:
        try:
            while True:
                if (tag == self._tagsStack[-1]):
                    # Its corresponding starting tag is found...
                    self._level -= 1
                    if self._isLastOpen and self._lastTag == tag:
                        next_line = False
                    else:
                        next_line = True
                    self._append_tag(f'</{tag}>', next_line=next_line)
                    self._tagsStack.pop()
                    self._lastTag = tag
                    self._isLastOpen = False
                    return
                elif (self._tagsStack[-1] in _SINGLETON_ELEMS):
                    self._append_tag(f'</{tag}>')
                    self._tagsStack.pop()
                    self._level -= 1
                # So far we have not found a corresponding starting tag for it
                elif tag in _SINGLETON_ELEMS:
                    return
                else:
                    raise TagMismatchError()
        except IndexError:
            # Tags stack is exhausted, so it's an orphan ending tag...
            raise TagMismatchError()
